/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import report.CreatedMessagesReport;
import report.NDN_Report;
import report.NamedDataReport;
import report.writeReportToFile;
import routing.NDN.DataReport;
import routing.NDN.Interest;
import routing.NDN.InterestReport;
import routing.NDN.socialTie;
import routing.PeopleShare.InterestList;
import core.Connection;
import core.Message;
import core.DTNHost;
import core.Settings;
import core.SimClock;


@SuppressWarnings("unused")
public class NDNRouter extends ActiveRouter {
	
	DataReport R;
	InterestReport I;

	/** PIFP router's setting namespace ({@value})*/ 
	public static final String NDN_NS = "NDNRouter";	
	public static final String INTEREST_PACKAGE = "pi";
	public static final String CONTENT_PACKAGE = "pc";
	
	public static final String SizeHigh = "sizeHigh";
	public static final String SizeLow = "sizeLow";

	/** delivery predictability initialization constant*/
	public static final double P_INIT = 0.75;
	/** delivery predictability transitivity scaling constant default value */
	public static final double DEFAULT_BETA = 0.25;
	/** delivery predictability aging constant */
	public static final double GAMMA = 0.98;
	
	/**
	 * Number of seconds in time unit -setting id ({@value}).
	 * How many seconds one time unit is when calculating aging of 
	 * delivery predictions. Should be tweaked for the scenario.*/
	public static final String SECONDS_IN_UNIT_S ="secondsInTimeUnit";
	
	/**
	 * Transitivity scaling constant (beta) -setting id ({@value}).
	 * Default value for setting is {@link #DEFAULT_BETA}.
	 */
	public static final String BETA_S = "beta";

	/** the value of nrof seconds in time unit -setting */
	protected int secondsInTimeUnit;
	/** value of beta setting */
	protected double beta;
	
	/** Delivery predictability */
	private Map<DTNHost, Double> preds;
	
	/** Last delivery predictability update (sim)time */
	protected double lastAgeUpdate;
	
	/** The list of all known interests*/
	protected Map<String, InterestList> li;
	
	/** Face list where arrived interests*/
	protected Map<String, ArrayList<DTNHost>> lei;
	
	/** Forwarding Information Base*/
	protected HashMap<String, Double> FIB;
	
	/** Pending Interest Table*/
	protected HashMap<String, PITEntries> PIT;
	
	/** Content Store*/
	protected Map<String, Integer> lc;
	
	
	/** PIT with Counting Bloom Filter*/
	CBloomFilter<String> bloomPIT;
	
	/** FIB with  Counting Bloom Filter*/
	CBloomFilter<String> bloomFIB;
	
	/** CS with  Counting Bloom Filter*/
	CBloomFilter<String> bloomLC;
	
	/** Auxiliary list of messages to remove */
	public Map<String, Double> toremove;
	
	double p_fp = 0.01; // False positive probability
	int n = 1000; // Expected number of elements
	int n_bitsforCounting = 4; // Number Bits for Couting
	
	
	private boolean hop = false;

	boolean LC = false; //LoadContent
	public static final String CONTENT = "content";
	public static final String HOSTS = "nrofHosts";
	public String  contentList = "";
	private int nrofHosts;
	private int id = 0;
	public int LowSize = 0;
	public int HighSize = 0;
	

	public int nrdatasent;
	public int nrinterestsreceived;
	public Double popularity;
	writeReportToFile wtf;
	
	
	int ttltoremove = 30;
	boolean isProducer;

	
	public  NDNRouter(Settings s) {
		super(s);		
		Settings NDNRouterSettings= new Settings(NDN_NS);
		contentList = NDNRouterSettings.getSetting(CONTENT);
		nrofHosts = NDNRouterSettings.getInt(HOSTS);
		HighSize = NDNRouterSettings.getInt(SizeHigh);
		LowSize = NDNRouterSettings.getInt(SizeLow);
		HighSize= HighSize*1000;
		LowSize= LowSize*1000;
	
		if (NDNRouterSettings.contains(BETA_S)) {
			beta = NDNRouterSettings.getDouble(BETA_S);
		}
		else {
			beta = DEFAULT_BETA;
		}
		
		initRouter();
	}
	
	protected NDNRouter(NDNRouter r) {
		super(r);
		this.contentList = r.contentList ;
		this.nrofHosts = r.nrofHosts;
		this.secondsInTimeUnit = r.secondsInTimeUnit;
		this.beta = r.beta;
		this.LowSize = r.LowSize;
		this.HighSize = r.HighSize;
		
		initRouter();
	}
	
	
	/** Init PIT, FIB, LC*/
	public void initRouter(){
		
		this.li = new HashMap<String, InterestList>();
		nrdatasent = 0;
		nrinterestsreceived = 0;
		this.popularity = 0.0;
		this.isProducer = false;
		
		//Create a Counting Bloom filter that has a FP rate of 0.01 when 1000 are inserted
		//and uses 4 Bits for Counting
		this.bloomPIT = new CBloomFilter<>(n, p_fp, n_bitsforCounting);
		this.bloomFIB = new CBloomFilter<>(n, p_fp, n_bitsforCounting);
		this.bloomLC = new CBloomFilter<>(n, p_fp, n_bitsforCounting);

		this.lei = new HashMap<String, ArrayList<DTNHost>>();
		this.FIB = new HashMap<String, Double>();
		
		this.lc = new HashMap<String, Integer>();
		
		this.PIT = new HashMap<String,PITEntries>();
		
		R = DataReport.getInstance();
		I = InterestReport.getInstance();
		
		wtf = writeReportToFile.getInstance();
		
		toremove = new HashMap<String, Double>();
		
		/*		
		if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
			loadRndContent();
		}*/
		
	}
		

	/**
	 * Load content from file and insert it in ContentStore of respective node
	 * 
	 */
	private void loadContent(){		
		String line;
		String[] contents, path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			
			while(br.ready()){		
				line = br.readLine();
				contents = line.split(" "); //contents[0] - host ; contents[1] - content
				path = contents[1].split("/"); 
				if(this.getHost().getAddress() == Integer.valueOf(contents[0])){
					
					if (!this.bloomLC.contains(contents[1])){
						this.lc.put(contents[1], 1);
						this.bloomLC.add(contents[1]);
						this.isProducer = true;
					}

					if(NDN_Report.list_of_publishers.containsValue(path[0])){
						continue;
					}
					else{
						NDN_Report.list_of_publishers.put(Integer.valueOf(contents[0]), path[0]);
					}
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	
	
/*	private void loadRndContent(){		
		String line;
		String[] path;
		boolean temp = false;
		int rndHost = getRandomHost();
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			while(br.ready()){		
				line = br.readLine();
				path = line.split("/");
				if(NDN_Report.list_of_publishers.containsValue(path[0])){
					continue;
				}
				else{
					do{
						if(NDN_Report.list_of_publishers.containsKey(rndHost)){
							rndHost = getRandomHost();
					}
						else{
							temp = true;
							NDN_Report.list_of_publishers.put(rndHost, path[0]);
						}
					}while(temp == false);
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}
	*/
	
	
	/**
	 * Load content from file 
	 * 
	 */
	private void uploadContent(){
		String line;
		String[] path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			int n = 0;
			if(NDN_Report.list_of_publishers.containsKey(this.getHost().getAddress())){
				String domain = NDN_Report.list_of_publishers.get(this.getHost().getAddress());
				while(br.ready()){		
					line = br.readLine();
					path = line.split("/"); 					
					if(domain.equalsIgnoreCase(path[0])){
						//addInterest("I" + n, line, 0, false, line);
						n++;			
					}
				}
				br.close();
				fr.close();
			}

		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	/**
	 * Generate a random identifier
	 */
	public int getRandomHost(){
		Random r = new Random();
		int n = r.nextInt(nrofHosts + 1);
		return n;
	}
	
	
	/**
	 * Create new entry in CS
	 * @param name of the content
	 * @param m is the message of content
	 * @param state, state (1 - is subscriver, 2 - is not subscriver)
	 */
	public void createCSEntrie(String name, Message m, int state)
	{
		this.bloomLC.add(name);
		this.lc.put(name, state); // adiciona a CS o conteudo
	}
	
	
	

	/**
	 * Create new entrie in PIT
	 * @param name name of the content
	 * @param h ArrayList of faces that sent interest
	 * @param t time of creation of the interest
	 * @param timetolive TTL of entry
	 * @param isSub indicates if the node is a subscriber or just carries interest
	 */
	public void createPITEntrie(String name,ArrayList<DTNHost> h, double t, double timetolive, boolean isSub)
	{
		String id = name;
		PITEntries entrie = new PITEntries();
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();
		
		if(!this.PIT.containsKey(name))
		{
			entrie = new PITEntries(id, h, t, timetolive, isSub);
			this.PIT.put(name,entrie);
			
		}
	}
	
	/**
	 * Delete PIT entry
	 * @param name of the content
	 */
	public void deletePITEntrie(String name)
	{
		bloomPIT.remove(name);
		this.PIT.remove(name);
	}
	
	
	/**
	 * Delete CS entry
	 * @param name, name of the content
	 */
	public void deleteCSEntrie(String name)
	{
		this.bloomLC.remove(name);
		this.lc.remove(name);
	}

		
	/**
	 * Generates a (random) message size
	 * @return message size
	 */
	protected int drawMessageSize() {
	
		Random r = new Random();
		int sizeRangeHigh = HighSize; // avg musica
		int sizeRangeLow = LowSize; //avg webpages
		int sizeDiff = sizeRangeLow == sizeRangeHigh ? 0 : 
			r.nextInt(sizeRangeHigh - sizeRangeLow);
		
		return sizeRangeLow + sizeDiff;
	}
	
	
	
	
	public int getMessageLenght(Message m, String data){
		int lenght = m.getFrom().toString().length() + m.getTo().toString().length() + data.length();
		return lenght;
	}
	
	/**
	 * Called when a connection's state changes. This version doesn't do 
	 * anything but subclasses may want to override this.
	 */
	@Override
	public void changedConnection(Connection con) { }
	
	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		return m;
	}
	

		
	/** Insert new interest in BloomPIT, PIT
	 * Define if this router is subscriber or not
	 * @param m, message Interest
	 * @param isSub, true if this router is subscriber, false if isn't subscriber
	 * @param facehost, EID of the host
	 * 
	 * */
	public void insertPI(Message m, boolean isSub, DTNHost interfacehost)
	{
		System.out.println("from: "+ m.getFrom()  + " TO: " + m.getTo());
		ArrayList<DTNHost> newList = new ArrayList<DTNHost>();

		String data = m.getData(); 
		DTNHost host = m.getFrom(); 
		String idInterest = m.getId();
		String[] datas = data.split(" ");
		String name = datas[1];
	
		if(this.bloomPIT.contains(name)) // if it exists at PIT add only the identifier of the node that requested
		{ 
			if(!this.PIT.get(name).getAllHosts().contains(host))
			{
				this.PIT.get(name).addHost(host);
				if(this.lei.containsKey(name))
				{
					if(!this.lei.get(name).contains(host)) this.lei.get(name).add(host);
					if(interfacehost!=null)
					{
						if(!this.lei.get(name).contains(interfacehost)) this.lei.get(name).add(interfacehost);
					}
				}
			}
						
			if(!I.LI.containsKey(idInterest))
			{
				I.createNewRinterest(idInterest, name);
			}
			updatePIT_TTL(m, name);
		}
		else //There is no entry in the PIT, create a new entry
		{ 			
			if(!I.LI.containsKey(idInterest))
			{
				I.createNewRinterest(idInterest, name);
			}
			if(this.bloomPIT.add(name)) 
			{ 				
				System.out.println("Insert interest  " + (String) name + " em " + m.getFrom()  );
				newList.add(host);
				if(interfacehost!=null)newList.add(interfacehost);
				createPITEntrie(name, newList, SimClock.getTime(), m.getTtl(), isSub); // cria nova entrada na PIT
				this.lei.put(name, newList);
			}
		}
	}
	
	/** 
	 * This method removes entries from CS
	 * @param ArrayList<String> messages, is a list of messages to remove 
	 * */
	public void updatelc(ArrayList<String> messages)
	{
		for (String name : messages)
		{
			deleteCSEntrie(name); // delete entry
			if(lei.containsKey(name)) this.lei.remove(name); // remove entry, even if you have not reached to the final consumer
		}
	}
	
	

	
	/** 
	 * This method allows the management of CS 
	 * The entries corresponding to the message with the TTL expired are deleted
	 * If the node is a producer, the messages are not deleted 
	 * If the node is a final consumer(if the subscriber has been of the interest),
	 * the messages are not deleted 
	 * */
	public void bufferman()
	{
		Collection<Message> msgCollection = getMessageCollection();
		ArrayList<String> messages = new ArrayList<String>();
			
		for (String name : this.lc.keySet())
		{
			int flag = -1;
			if((checkifExistesDeliveredmessages(name)==true) || (this.isProducer == true)) continue; //if is a final consumer or is a producer
			else
			{
				for (Message m : msgCollection) 
				{
					String data = m.getData();
					String datas[] = data.split(" ");
					String key = datas[1];
						
					if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE))
					{	
						if(m.getData().split(" ")[1].equalsIgnoreCase(name))
						{	
							if(m.getTtl()>0)
							{
								flag = 1;
								return;
							}
						}
					}
				}
				if(flag!=1)
				{
					messages.add(name);
				}
			}
		}
			
		if(!messages.isEmpty())
		{
			updatelc(messages); 
		}
			
		for(String k : this.toremove.keySet())
		{
			int TTL =  (int)( ((this.ttltoremove * 60) -
						(SimClock.getTime()-this.toremove.get(k))) /60.0 );
			if(TTL <= 0) deleteMessage(k, true);
		}
			
	}
		
		
		
	/** 
	 * Drop messages of Interest and Data
	 * TTL is updated and if it expired messages are removed
	 * The TTL of the PIT is adjusted with the arrival of new interests
	 * */	
	@Override
	protected void dropExpiredMessages() 
	{	
		Message[] messages = getMessageCollection().toArray(new Message[0]);
		ArrayList<String> messagestodelete = new ArrayList<String>();
		ArrayList<String> ListofMessages = new ArrayList<String>();
				
		for (int i=0; i<messages.length; i++) 
		{
			int ttl = messages[i].getTtl(); 

			if (ttl <= 0) 
			{	
				String data =  messages[i].getData();
				String datas[] = data.split(" ");
				String key = datas[1];
						
				if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) //"pi"
				{
					if (this.bloomPIT.contains(key))
					{ 	
						int val = updatePIT_TTL(messages[i],  key); //TTL adjust in PIT
						if (val> 0)continue;
						else 
						{		
							deleteMessage(messages[i].getId(), true);
							deletePITEntrie(key);
							
						}
					}
				}
				else if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)) // "pc"
				{
					if(this.bloomLC.contains(key))
					{
						if( messages[i].getTtl() <=0)  // TTL expired
						{
							deleteMessage(messages[i].getId(), true); // remove data Message
							bufferman(); // Update ContentStore
						}
					}
				}
	
			}
					
		}

	}
		
	

	/** 
	* Soften the TTL value in the PIT
	* With the arrival of messages of interest, larger than the current TTL in PIT
	* difference = TTL message - TTL PIT
	* TTL PIT = difference + TTLPit
	* */	
	public int updatePIT_TTL(Message m, String Key)
	{
	  int PIT_TTL =  (int)( ((this.PIT.get(Key).ttl * 60) -
				(SimClock.getTime()-this.PIT.get(Key).timeCreated)) /60.0 ); // get PIT TTL
	  
	  int TTLMessage = m.getTtl(); // get Message TTL
	  
	  if(TTLMessage > PIT_TTL)
	  {
		  int  ttlToPIT = TTLMessage - PIT_TTL;
		  this.PIT.get(Key).ttl =  PIT_TTL + ttlToPIT; // set new TTL to PIT
	  }
	 
	  return (int)( ((this.PIT.get(Key).ttl * 60) -
				(SimClock.getTime()-this.PIT.get(Key).timeCreated)) /60.0 );

	}
	
	
	/** 
	*This method checks if the message exists in the repository("Repo")
	* @param name,  this String Identifies the name of content
	* @return boolean, true if the content exists and false in negative case
	* */	
	public boolean checkifExistesDeliveredmessages(String name)
	{
		for(String key : this.deliveredMessages.keySet())
		{
			
			if (this.deliveredMessages.get(key).getData().split(" ")[1].equalsIgnoreCase(name))
			{
				return true;
			}
		}
		return false;
	}
	
	
	@Override
	public void update() {
		super.update();
		if(!LC){
			LC=true;
			if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
				uploadContent();
			}
			else
				loadContent();			
		}
		wtf.writeNumberInterestsSastified(I, R);
		wtf.AvgDelaywithCI(I);
		wtf.AvgDelaywithCIAlternative(I);
	}
	
	@Override
	public NDNRouter replicate() {
		return new NDNRouter(this);
	}
	

}
